import io
import json
import os
import pathlib
import sys

import click
import httpx
from nanollm import user_dir
from tango.nanoLLM.nanollm.plugin_spec import hookimpl
from tango.nanoLLM.nanollm.models import Model, Options

try:
    from pydantic import Field, field_validator  # type: ignore
except ImportError:
    from pydantic.class_validators import \
        validator as field_validator  # type: ignore [no-redef]
    from pydantic.fields import Field

try:
    from llama_cpp import Llama
except ImportError:
    Llama = None
    print(
        "llama_cpp not installed, install with: pip install llama-cpp-python",
        file=sys.stderr,
    )

DEFAULT_LLAMA2_CHAT_SYSTEM_PROMPT = """
You are a helpful, respectful and honest assistant. Always answer as helpfully as possible, while being safe.  Your answers should not include any harmful, unethical, racist, sexist, toxic, dangerous, or illegal content. Please ensure that your responses are socially unbiased and positive in nature.

If a question does not make any sense, or is not factually coherent, explain why instead of answering something not correct. If you don't know the answer to a question, please don't share false information.
""".strip()


def _ensure_models_dir():
    directory = user_dir() / "llama-cpp" / "models"
    directory.mkdir(parents=True, exist_ok=True)
    return directory


def _ensure_models_file():
    directory = user_dir() / "llama-cpp"
    directory.mkdir(parents=True, exist_ok=True)
    filepath = directory / "models.json"
    if not filepath.exists():
        filepath.write_text("{}")
    return filepath

class LlamaCppPlugin:
    name = "llama_cpp"

    @hookimpl
    def register_models(self, register):
        directory = user_dir() / "llama-cpp"
        models_file = directory / "models.json"
        if not models_file.exists():
            return
        models = json.loads(models_file.read_text())
        for model_id, details in models.items():
            register(
                LlamaModel(
                    model_id,
                    details["path"],
                    is_llama2_chat=details.get("is_llama2_chat", False),
                ),
                aliases=details["aliases"],
            )
    
    @hookimpl
    def register_commands(self, cli):
        @cli.group()
        def llama_cpp():
            "Commands for registering llama.cpp GGML models with LLM"
    
        @llama_cpp.command()
        def models_file():
            "Display the path to the models.json file"
            directory = user_dir() / "llama-cpp"
            directory.mkdir(parents=True, exist_ok=True)
            models_file = directory / "models.json"
            click.echo(models_file)
    
        @llama_cpp.command()
        def models_dir():
            "Display the path to the directory holding downloaded models"
            click.echo(_ensure_models_dir())
    
        @llama_cpp.command()
        @click.argument("url")
        @click.option(
            "aliases",
            "-a",
            "--alias",
            multiple=True,
            help="Alias(es) to register the model under",
        )
        @click.option(
            "--llama2-chat",
            is_flag=True,
            help="Mark as using the Llama 2 chat prompt format",
        )
        def download_model(url, aliases, llama2_chat):
            "Download and register a model from a URL"
            with httpx.stream("GET", url, follow_redirects=True) as response:
                total_size = response.headers.get("content-length")
    
                filename = url.split("/")[-1]
                download_path = _ensure_models_dir() / filename
                if download_path.exists():
                    raise click.ClickException(f"File already exists at {download_path}")
    
                with open(download_path, "wb") as fp:
                    if total_size is not None:  # If Content-Length header is present
                        total_size = int(total_size)
                        with click.progressbar(
                            length=total_size,
                            label="Downloading {}".format(human_size(total_size)),
                        ) as bar:
                            for data in response.iter_bytes(1024):
                                fp.write(data)
                                bar.update(len(data))
                    else:  # If Content-Length header is not present
                        for data in response.iter_bytes(1024):
                            fp.write(data)
    
                click.echo(f"Downloaded model to {download_path}", err=True)
                models_file = _ensure_models_file()
                models = json.loads(models_file.read_text())
                model_id = download_path.stem
                info = {
                    "path": str(download_path.resolve()),
                    "aliases": aliases,
                }
                if llama2_chat:
                    info["is_llama2_chat"] = True
                models[model_id] = info
                models_file.write_text(json.dumps(models, indent=2))
    
        @llama_cpp.command()
        @click.argument(
            "filepath", type=click.Path(exists=True, dir_okay=False, resolve_path=True)
        )
        @click.option(
            "aliases",
            "-a",
            "--alias",
            multiple=True,
            help="Alias(es) to register the model under",
        )
        @click.option(
            "--llama2-chat",
            is_flag=True,
            help="Mark as using the Llama 2 chat prompt format",
        )
        def add_model(filepath, aliases, llama2_chat):
            "Register a GGML model you have already downloaded with LLM"
            models_file = _ensure_models_file()
            models = json.loads(models_file.read_text())
            path = pathlib.Path(filepath)
            model_id = path.stem
            info = {
                "path": str(path.resolve()),
                "aliases": aliases,
            }
            if llama2_chat:
                info["is_llama2_chat"] = True
            models[model_id] = info
            models_file.write_text(json.dumps(models, indent=2))
    
        @llama_cpp.command()
        def models():
            "List registered GGML models"
            models_file = _ensure_models_file()
            models = json.loads(models_file.read_text())
            click.echo(json.dumps(models, indent=2))


class LlamaModel(Model):
    can_stream = True

    class ModelOptions(Options):
        verbose: bool = Field(
            description="Whether to print verbose output from the model", default=False
        )
        no_gpu: bool = Field(
            description="Remove the default n_gpu_layers=1 argument", default=False
        )
        n_ctx: int = Field(description="n_ctx argument, defaults to 4000", default=None)

    def __init__(self, model_id, path, is_llama2_chat: bool = False):
        self.model_id = model_id
        self.path = path
        self.is_llama2_chat = is_llama2_chat
        self.default_system_prompt = None

    def build_llama2_chat_prompt(self, prompt, conversation):
        prompt_bits = []
        # First figure out the system prompt
        system_prompt = None
        if prompt.system:
            system_prompt = prompt.system
        else:
            # Look for a system prompt in the conversation
            if conversation is not None:
                for prev_response in conversation.responses:
                    if prev_response.prompt.system:
                        system_prompt = prev_response.prompt.system
        if system_prompt is None:
            system_prompt = (
                self.default_system_prompt or DEFAULT_LLAMA2_CHAT_SYSTEM_PROMPT
            )

        # Now build the prompt pieces
        first = True
        if conversation is not None:
            for prev_response in conversation.responses:
                prompt_bits.append("<s>[INST] ")
                if first:
                    prompt_bits.append(
                        f"<<SYS>>\n{system_prompt}\n<</SYS>>\n\n",
                    )
                first = False
                prompt_bits.append(
                    f"{prev_response.prompt.prompt} [/INST] ",
                )
                prompt_bits.append(
                    f"{prev_response.text()} </s>",
                )

        # Add the latest prompt
        if not prompt_bits:
            # Start with the system prompt
            prompt_bits.append("<s>[INST] ")
            prompt_bits.append(
                f"<<SYS>>\n{system_prompt}\n<</SYS>>\n\n",
            )
        else:
            prompt_bits.append("<s>[INST] ")
        prompt_bits.append(f"{prompt.prompt} [/INST] ")
        return prompt_bits

    def execute(self, prompt, stream, response, conversation):
        with SuppressOutput(verbose=prompt.options.verbose):
            kwargs = {"n_ctx": prompt.options.n_ctx or 4000, "n_gpu_layers": 1}
            if prompt.options.no_gpu:
                kwargs.pop("n_gpu_layers")
            llm_model = Llama(
                model_path=self.path, verbose=prompt.options.verbose, **kwargs
            )
            if self.is_llama2_chat:
                prompt_bits = self.build_llama2_chat_prompt(prompt, conversation)
                prompt_text = "".join(prompt_bits)
                response._prompt_json = {"prompt_bits": prompt_bits}
            else:
                prompt_text = prompt.prompt
            stream = llm_model(prompt_text, stream=True, max_tokens=4000)
            for item in stream:
                # Each item looks like this:
                # {'id': 'cmpl-00...', 'object': 'text_completion', 'created': .., 'model': '/path', 'choices': [
                #   {'text': '\n', 'index': 0, 'logprobs': None, 'finish_reason': None}
                # ]}
                yield item["choices"][0]["text"]


def human_size(num_bytes):
    """Return a human readable byte size."""
    for unit in ["B", "KB", "MB", "GB", "TB", "PB"]:
        if num_bytes < 1024.0:
            break
        num_bytes /= 1024.0
    return f"{num_bytes:.2f} {unit}"


class SuppressOutput:
    def __init__(self, verbose: bool = False):
        self.verbose = verbose

    def __enter__(self):
        if self.verbose:
            return
        # Save a copy of the current file descriptors for stdout and stderr
        self.stdout_fd = os.dup(1)
        self.stderr_fd = os.dup(2)

        # Open a file to /dev/null
        self.devnull_fd = os.open(os.devnull, os.O_WRONLY)

        # Replace stdout and stderr with /dev/null
        os.dup2(self.devnull_fd, 1)
        os.dup2(self.devnull_fd, 2)

        # Writes to sys.stdout and sys.stderr should still work
        self.original_stdout = sys.stdout
        self.original_stderr = sys.stderr
        sys.stdout = os.fdopen(self.stdout_fd, "w")
        sys.stderr = os.fdopen(self.stderr_fd, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.verbose:
            return
        # Restore stdout and stderr to their original state
        os.dup2(self.stdout_fd, 1)
        os.dup2(self.stderr_fd, 2)

        # Close the saved copies of the original stdout and stderr file descriptors
        os.close(self.stdout_fd)
        os.close(self.stderr_fd)

        # Close the file descriptor for /dev/null
        os.close(self.devnull_fd)

        # Restore sys.stdout and sys.stderr
        sys.stdout = self.original_stdout
        sys.stderr = self.original_stderr

PLUGIN_CLASS = LlamaCppPlugin


import pydantic
from pydantic import BaseModel
import string
from typing import Optional, Any, Dict, List, Tuple
import click
import yaml
from nanollm import user_dir
from sqlite_utils import Database
from tango.configs.dbs_config import NANOLLM_DB

def template_dir():
    path = user_dir() / "templates"
    path.mkdir(parents=True, exist_ok=True)
    return path

def load_template(name):
    path = template_dir() / f"{name}.yaml"
    if not path.exists():
        raise click.ClickException(f"Invalid template: {name}")
    try:
        loaded = yaml.safe_load(path.read_text())
    except yaml.YAMLError as ex:
        raise click.ClickException("Invalid YAML: {}".format(str(ex)))
    if isinstance(loaded, str):
        return Template(name=name, prompt=loaded)
    loaded["name"] = name
    try:
        return Template(**loaded)
    except pydantic.ValidationError as ex:
        msg = "A validation error occurred:\n"
        raise click.ClickException(msg)
    
class Template(BaseModel):
    name: str
    prompt: Optional[str] = None
    system: Optional[str] = None
    model: Optional[str] = None
    defaults: Optional[Dict[str, Any]] = None

    class Config:
        extra = "forbid"

    class MissingVariables(Exception):
        pass

    def evaluate(
        self, input: str, params: Optional[Dict[str, Any]] = None
    ) -> Tuple[str, Optional[str]]:
        params = params or {}
        params["input"] = input
        if self.defaults:
            for k, v in self.defaults.items():
                if k not in params:
                    params[k] = v

        # Extract placeholders from the template
        expected_params = set(self.extract_vars(string.Template(self.prompt or "")))
        if self.system:
            expected_params.update(self.extract_vars(string.Template(self.system)))
        
        # Check for unexpected parameters
        unexpected_params = set(params.keys()) - expected_params
        unexpected_params.discard("input")
        if unexpected_params:
            raise ValueError(f"Unexpected parameters: {', '.join(unexpected_params)}")
    
        prompt: Optional[str] = None
        system: Optional[str] = None
        if not self.prompt:
            system = self.interpolate(self.system, params)
            prompt = input
        elif self.system:
            prompt = self.interpolate(self.prompt, params)
            system = self.interpolate(self.system, params)
        else:
            prompt = self.interpolate(self.prompt, params)

        return prompt, system

    @classmethod
    def interpolate(cls, text: Optional[str], params: Dict[str, Any]) -> Optional[str]:
        if not text:
            return text
        # Confirm all variables in text are provided
        string_template = string.Template(text)
        vars = cls.extract_vars(string_template)
        missing = [p for p in vars if p not in params]
        if missing:
            raise cls.MissingVariables(
                "Missing variables: {}".format(", ".join(missing))
            )
        return string_template.substitute(**params)

    @staticmethod
    def extract_vars(string_template: string.Template) -> List[str]:
        return [
            match.group("named")
            for match in string_template.pattern.finditer(string_template.template)
        ]
    
    def log_template(self):
        db = Database(NANOLLM_DB)
        db["templates"].upsert(self.dict(), pk="name")
        


# 
import click
import contextlib
import json
import os
import nanollm as llm
from pathlib import Path 
from pydantic import Field
import subprocess
import sys
import textwrap
from typing import Optional
from tango.nanoLLM.nanollm.plugin_spec import hookimpl
from tango.nanoLLM.nanollm.models import Model, Options

MODEL_URLS = {
    "Llama-2-7b-chat": "https://huggingface.co/mlc-ai/mlc-chat-Llama-2-7b-chat-hf-q4f16_1",
    "Llama-2-13b-chat": "https://huggingface.co/mlc-ai/mlc-chat-Llama-2-13b-chat-hf-q4f16_1",
    "Llama-2-70b-chat": "https://huggingface.co/mlc-ai/mlc-chat-Llama-2-70b-chat-hf-q4f16_1",
    "WizardCoder-15b-chat": "https://huggingface.co/mlc-ai/mlc-chat-WizardCoder-15B-V1.0-q4f16_1",
}

MLC_INSTALL = (
    "You must install mlc_chat first. "
    "See https://github.com/simonw/llm-mlc for instructions."
)


def is_git_lfs_command_available():
    try:
        subprocess.run(
            ["git", "lfs"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            check=True,
        )
        return True
    except subprocess.CalledProcessError:
        return False


def is_git_lfs_installed():
    try:
        # Run the git config command to get the filter value
        result = subprocess.check_output(
            ["git", "config", "--get", "filter.lfs.clean"], encoding="utf-8"
        ).strip()

        # If the result contains "git-lfs clean", it's likely that Git LFS is installed
        if "git-lfs clean" in result:
            return True
        else:
            return False
    except subprocess.CalledProcessError:
        # If the command fails, it's likely that the configuration option isn't set
        return False


def _ensure_models_dir():
    directory = llm.user_dir() / "llama-cpp" / "models"
    directory.mkdir(parents=True, exist_ok=True)
    return directory


def _ensure_models_file():
    directory = llm.user_dir() / "llama-cpp"
    directory.mkdir(parents=True, exist_ok=True)
    filepath = directory / "models.json"
    if not filepath.exists():
        filepath.write_text("{}")
    return filepath

class MLCPlugin:
    name = "mlc"

    @hookimpl
    def register_models(self, register):
        directory = llm.user_dir() / "mlc"
        models_dir = directory / "dist" / "prebuilt"
        if not models_dir.exists():
            print('models_dir does not exist', models_dir)
            return
        for child in models_dir.iterdir():
            if child.is_dir() and child.name != "lib":
                # It's a model! Register it
                register(
                    MlcModel(
                        model_id=child.name,
                        model_path=str(child.absolute()),
                    )
                )
    @staticmethod 
    def download_model(url, aliases):
        if url in MODEL_URLS:
            # Set that up as an alias too
            aliases = list(aliases) + [url]
        if not url.startswith("https://"):
            raise click.BadParameter("Invalid model name or URL")
        directory = llm.user_dir() / "mlc"
        prebuilt_dir = directory / "dist" / "prebuilt"
        if not prebuilt_dir.exists():
            raise click.ClickException("You must run 'nllm mlc setup' first")
        # Run git clone URL dist/prebuilt
        last_bit = url.split("/")[-1]
        git_clone_command = [
            "git",
            "clone",
            url,
            str((prebuilt_dir / last_bit).absolute()),
        ]
        print('run sub process', git_clone_command)
        subprocess.run(git_clone_command, check=True)
        if aliases:
            aliases_path = llm.user_dir() / "aliases.json"
            if not aliases_path.exists():
                aliases_path.write_text("{}")
            aliases_data = json.loads(aliases_path.read_text())
            for alias in aliases:
                aliases_data[alias] = last_bit
            aliases_path.write_text(json.dumps(aliases_data, indent=2))
 
    @hookimpl
    def register_commands(self, cli):
        @cli.group()
        def mlc():
            "Commands for managing MLC models"
    
        @mlc.command()
        def setup():
            "Finish setting up MLC, step by step"
            directory = llm.user_dir() / "mlc"
            directory.mkdir(parents=True, exist_ok=True)
            if not is_git_lfs_command_available():
                raise click.ClickException(
                    "Git LFS is not installed. You should run 'brew install git-lfs' or similar."
                )
            if not is_git_lfs_installed():
                click.echo(
                    "Git LFS is not installed. Should I run 'git lfs install' for you?"
                )
                if click.confirm("Install Git LFS?"):
                    subprocess.run(["git", "lfs", "install"], check=True)
                else:
                    raise click.ClickException(
                        "Git LFS is not installed. You should run 'git lfs install'."
                    )
            # Now we have git-lfs installed, ensure we have cloned the repo
            dist_dir = directory / "dist"
            if not dist_dir.exists():
                click.echo("Downloading prebuilt binaries...")
                # mkdir -p dist/prebuilt
                (dist_dir / "prebuilt").mkdir(parents=True, exist_ok=True)
                # git clone
                git_clone_command = [
                    "git",
                    "clone",
                    "https://github.com/mlc-ai/binary-mlc-llm-libs.git",
                    str((dist_dir / "prebuilt" / "lib").absolute()),
                ]
                subprocess.run(git_clone_command, check=True)
            click.echo("Ready to install models in {}".format(directory))
            # Do we have mlc_chat installed?
            try:
                import mlc_chat
            except ImportError:
                raise click.ClickException(MLC_INSTALL)
    
        @mlc.command(
            help=textwrap.dedent(
                """
            Download and register a model from a URL
    
            Try one of these names:
            
            \b
            {}
            """
            ).format("\n".join("- {}".format(key) for key in MODEL_URLS.keys()))
        )
        @click.argument("name_or_url")
        @click.option(
            "aliases", "-a", "--alias", multiple=True, help="Alias to use for this model"
        )
        def download_model(name_or_url, aliases):
            url = MODEL_URLS.get(name_or_url) or name_or_url
            self.download_model(url, aliases)
   
        @mlc.command()
        def models():
            "List installed MLC models"
    
            for model_alias in llm.get_models_with_aliases():
                if isinstance(model_alias.model, MlcModel):
                    aliases = ", ".join(model_alias.aliases)
                    if aliases:
                        aliases = " (aliases: {})".format(aliases)
                    click.echo(
                        "{}{}".format(
                            model_alias.model,
                            aliases,
                        )
                    )
    
        @mlc.command()
        def models_dir():
            "Display the path to the directory holding downloaded models"
            directory = llm.user_dir() / "mlc" / "dist" / "prebuilt"
            click.echo(directory.absolute())
    
        @mlc.command(
            context_settings={
                "ignore_unknown_options": True,
                "allow_extra_args": True,
            }
        )
        @click.pass_context
        def pip(ctx, **kwargs):
            "Run pip in the LLM virtual environment"
            cmd = [sys.executable, "-m", "pip"] + ctx.args
            subprocess.run(cmd)
    

class MlcModel(Model):
    can_stream = True

    class ModelOptions(Options):
        temperature: Optional[float] = Field(
            description=(
                "A higher temperature encourages more diverse outputs, while a "
                "lower temperature produces more deterministic outputs"
            ),
            ge=0,
            le=1,
            default=None,
        )
        top_p: Optional[float] = Field(
            description=(
                "At each step, we select tokens from the minimal set that has a "
                "cumulative probability exceeding this value"
            ),
            ge=0,
            le=1,
            default=None,
        )
        repetition_penalty: Optional[float] = Field(
            description=(
                "Controls the likelihood of the model generating repeated texts"
            ),
            ge=0,
            le=1,
            default=None,
        )
        max_gen_len: Optional[int] = Field(
            description="The maximum length of the output text",
            gt=0,
            default=None,
        )

    def __init__(self, model_id, model_path):
        self.model_id = model_id
        self.model_path = model_path
        self.chat_mod = None  # Lazy loading

    def execute(self, prompt, stream, response, conversation):
        try:
            import mlc_chat
            from mlc_chat.base import get_delta_message
            import mlc_chat.chat_module
        except ImportError:
            raise click.ClickException(MLC_INSTALL)

        # Disable print() in that module
        def noop(*args, **kwargs):
            pass

        mlc_chat.chat_module.__dict__["print"] = noop

        class StreamingChatModule(mlc_chat.ChatModule):
            def generate_iter(self, prompt):
                curr_message = ""
                self._prefill(prompt)
                while not self._stopped():
                    self._decode()
                    new_msg = self._get_message()
                    delta = get_delta_message(curr_message, new_msg)
                    curr_message = new_msg
                    yield delta

        with SuppressOutput():
            config_kwargs = {}

            system_prompt = None
            if conversation:
                messages = []
                # Populate messages from the conversation history
                for prev_response in conversation.responses:
                    if prev_response.prompt.system:
                        # Use the last set system prompt in that sequence
                        system_prompt = prev_response.prompt.system
                    messages.extend(
                        [
                            ["USER", prev_response.prompt.prompt],
                            ["ASSISTANT", prev_response.text()],
                        ]
                    )
                if messages:
                    config_kwargs["messages"] = messages
                    config_kwargs["offset"] = len(messages)

            if self.chat_mod is None:
                with temp_chdir(llm.user_dir() / "mlc"):
                    self.chat_mod = StreamingChatModule(model=self.model_path)

            if prompt.system:
                system_prompt = prompt.system

            if system_prompt is not None:
                config_kwargs["system"] = system_prompt

            response._prompt_json = config_kwargs if config_kwargs else None

            chat_config_kwargs = {
                "max_gen_len": prompt.options.max_gen_len or 4000,
                "conv_config": mlc_chat.ConvConfig(**config_kwargs),
            }
            if prompt.options.temperature is not None:
                chat_config_kwargs["temperature"] = prompt.options.temperature
            if prompt.options.top_p is not None:
                chat_config_kwargs["top_p"] = prompt.options.top_p
            if prompt.options.repetition_penalty is not None:
                chat_config_kwargs[
                    "repetition_penalty"
                ] = prompt.options.repetition_penalty

            self.chat_mod.reset_chat(mlc_chat.ChatConfig(**chat_config_kwargs))

            if stream:
                yield from self.chat_mod.generate_iter(prompt=prompt.prompt)
            else:
                # All in one go
                yield self.chat_mod.generate(prompt=prompt.prompt)


@contextlib.contextmanager
def temp_chdir(path):
    old_dir = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_dir)


class SuppressOutput:
    def __enter__(self):
        # Save a copy of the current file descriptors for stdout and stderr
        self.stdout_fd = os.dup(1)
        self.stderr_fd = os.dup(2)

        # Open a file to /dev/null
        self.devnull_fd = os.open(os.devnull, os.O_WRONLY)

        # Replace stdout and stderr with /dev/null
        os.dup2(self.devnull_fd, 1)
        os.dup2(self.devnull_fd, 2)

        # Writes to sys.stdout and sys.stderr should still work
        self.original_stdout = sys.stdout
        self.original_stderr = sys.stderr
        sys.stdout = os.fdopen(self.stdout_fd, "w")
        sys.stderr = os.fdopen(self.stderr_fd, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore stdout and stderr to their original state
        os.dup2(self.stdout_fd, 1)
        os.dup2(self.stderr_fd, 2)

        # Close the saved copies of the original stdout and stderr file descriptors
        os.close(self.stdout_fd)
        os.close(self.stderr_fd)

        # Close the file descriptor for /dev/null
        os.close(self.devnull_fd)

        # Restore sys.stdout and sys.stderr
        sys.stdout = self.original_stdout
        sys.stderr = self.original_stderr

# Expose the plugin class for dynamic loading
PLUGIN_CLASS = MLCPlugin



from tango.nanoLLM.nanollm.plugin_spec import hookimpl

class MockPlugin:
    """A mock plugin for testing."""
    
    @hookimpl
    def register_models(self, register):
        #print(f"MockPlugin's llm_load_model method invoked with register: {register}")
        pass




import pluggy

hookspec = pluggy.HookspecMarker("nanoLLM")
hookimpl = pluggy.HookimplMarker("nanoLLM")

class NanoLLMSpec:
    """Hook specifications for nanoLLM."""
    
    @hookspec
    def register_models(self, register):
        """Load a language model by name or path."""
        pass

    @hookspec
    def register_commands(self, cli):
        """Register additional CLI commands."""
        pass



import requests
import os
import openai
import click
try:
    from pydantic import field_validator, Field  # type: ignore
except ImportError:
    from pydantic.fields import Field
    from pydantic.class_validators import validator as field_validator  # type: ignore [no-redef]
from typing import Optional, Union, List
from tango.nanoLLM.nanollm.plugin_spec import hookimpl
from tango.nanoLLM.nanollm.models import Model, Options 

class OpenAIPlugin:
    name = "openai_chat_model"
    OPENAI_API_CHAT_URL = "https://api.openai.com/v1/chat/completions"
    HEADERS = {
        "Authorization": f"Bearer {os.environ.get('OPENAI_API_KEY')}",
        "Content-Type": "application/json"
    }
    
    @hookimpl
    def register_models(self, register):
        # For this plugin, loading is essentially ensuring we have an API key.
        # You can add more logic here if needed.
        if not os.environ.get('OPENAI_API_KEY'):
            raise ValueError("OPENAI_API_KEY environment variable not set.")
        register(Chat("gpt-3.5-turbo"), aliases=("3.5", "chatgpt"))
        register(Chat("gpt-3.5-turbo-16k"), aliases=("chatgpt-16k", "3.5-16k"))
        register(Chat("gpt-4"), aliases=("4", "gpt4"))
        register(Chat("gpt-4-32k"), aliases=("4-32k",))
        register(
            Completion("gpt-3.5-turbo-instruct", default_max_tokens=256),
            aliases=("3.5-instruct", "chatgpt-instruct"),
    )

    def predict(self, prompt: str) -> str:

        payload = {
            "model": "gpt-3.5-turbo",
            "messages": [{"role": "user", "content": prompt}],
            "max_tokens": 150
        }
        response = requests.post(self.OPENAI_API_CHAT_URL, headers=self.HEADERS, json=payload)

        if response.status_code != 200:
            return f"Error {response.status_code}: {response.json().get('error', {}).get('message', 'Unknown error')}"

        response_content = response.json().get("choices", [{}])[0].get("message", {}).get("content", "").strip()
        return response_content

class Chat(Model):
    needs_key = "openai"
    key_env_var = "OPENAI_API_KEY"
    can_stream: bool = True

    default_max_tokens = None

    class ModelOptions(Options):
        temperature: Optional[float] = Field(
            description=(
                "What sampling temperature to use, between 0 and 2. Higher values like "
                "0.8 will make the output more random, while lower values like 0.2 will "
                "make it more focused and deterministic."
            ),
            ge=0,
            le=2,
            default=None,
        )
        max_tokens: Optional[int] = Field(
            description="Maximum number of tokens to generate.", default=None
        )
        top_p: Optional[float] = Field(
            description=(
                "An alternative to sampling with temperature, called nucleus sampling, "
                "where the model considers the results of the tokens with top_p "
                "probability mass. So 0.1 means only the tokens comprising the top "
                "10% probability mass are considered. Recommended to use top_p or "
                "temperature but not both."
            ),
            ge=0,
            le=1,
            default=None,
        )
        frequency_penalty: Optional[float] = Field(
            description=(
                "Number between -2.0 and 2.0. Positive values penalize new tokens based "
                "on their existing frequency in the text so far, decreasing the model's "
                "likelihood to repeat the same line verbatim."
            ),
            ge=-2,
            le=2,
            default=None,
        )
        presence_penalty: Optional[float] = Field(
            description=(
                "Number between -2.0 and 2.0. Positive values penalize new tokens based "
                "on whether they appear in the text so far, increasing the model's "
                "likelihood to talk about new topics."
            ),
            ge=-2,
            le=2,
            default=None,
        )
        stop: Optional[str] = Field(
            description=("A string where the API will stop generating further tokens."),
            default=None,
        )
        logit_bias: Optional[Union[dict, str]] = Field(
            description=(
                "Modify the likelihood of specified tokens appearing in the completion. "
                'Pass a JSON string like \'{"1712":-100, "892":-100, "1489":-100}\''
            ),
            default=None,
        )

        @field_validator("logit_bias")
        def validate_logit_bias(cls, logit_bias):
            if logit_bias is None:
                return None

            if isinstance(logit_bias, str):
                try:
                    logit_bias = json.loads(logit_bias)
                except json.JSONDecodeError:
                    raise ValueError("Invalid JSON in logit_bias string")

            validated_logit_bias = {}
            for key, value in logit_bias.items():
                try:
                    int_key = int(key)
                    int_value = int(value)
                    if -100 <= int_value <= 100:
                        validated_logit_bias[int_key] = int_value
                    else:
                        raise ValueError("Value must be between -100 and 100")
                except ValueError:
                    raise ValueError("Invalid key-value pair in logit_bias dictionary")

            return validated_logit_bias

    def __init__(
        self,
        model_id,
        key=None,
        model_name=None,
        api_base=None,
        api_type=None,
        api_version=None,
        api_engine=None,
        headers=None,
    ):
        self.model_id = model_id
        self.key = key
        self.model_name = model_name
        self.api_base = api_base
        self.api_type = api_type
        self.api_version = api_version
        self.api_engine = api_engine
        self.headers = headers

    def __str__(self):
        return "OpenAI Chat: {}".format(self.model_id)

    def execute(self, prompt, stream=None, response=None, conversation=None):
        messages = []
        current_system = None
        if conversation is not None:
            for prev_response in conversation.responses:
                if (
                    prev_response.prompt.system
                    and prev_response.prompt.system != current_system
                ):
                    messages.append(
                        {"role": "system", "content": prev_response.prompt.system}
                    )
                    current_system = prev_response.prompt.system
                messages.append(
                    {"role": "user", "content": prev_response.prompt.prompt}
                )
                messages.append({"role": "assistant", "content": prev_response.text()})
        if prompt.system and prompt.system != current_system:
            messages.append({"role": "system", "content": prompt.system})
        messages.append({"role": "user", "content": prompt.prompt})
        response._prompt_json = {"messages": messages}
        kwargs = self.build_kwargs(prompt)
        if stream:
            completion = openai.ChatCompletion.create(
                model=self.model_name or self.model_id,
                messages=messages,
                stream=True,
                **kwargs,
            )
            chunks = []
            for chunk in completion:
                chunks.append(chunk)
                content = chunk["choices"][0].get("delta", {}).get("content")
                if content is not None:
                    yield content
            response.response_json = combine_chunks(chunks)
        else:
            completion = openai.ChatCompletion.create(
                model=self.model_name or self.model_id,
                messages=messages,
                stream=False,
                **kwargs,
            )
            response.response_json = completion.to_dict_recursive()
            yield completion.choices[0].message.content

    def build_kwargs(self, prompt):
        kwargs = dict(not_nulls(prompt.options))
        if "max_tokens" not in kwargs and self.default_max_tokens is not None:
            kwargs["max_tokens"] = self.default_max_tokens
        if self.api_base:
            kwargs["api_base"] = self.api_base
        if self.api_type:
            kwargs["api_type"] = self.api_type
        if self.api_version:
            kwargs["api_version"] = self.api_version
        if self.api_engine:
            kwargs["engine"] = self.api_engine
        if self.needs_key:
            if self.key:
                kwargs["api_key"] = self.key
        else:
            # OpenAI-compatible models don't need a key, but the
            # openai client library requires one
            kwargs["api_key"] = "DUMMY_KEY"
        if self.headers:
            kwargs["headers"] = self.headers
        return kwargs


class Completion(Chat):
    class ModelOptions(Chat.ModelOptions):
        logprobs: Optional[int] = Field(
            description="Include the log probabilities of most likely N per token",
            default=None,
            le=5,
        )

    def __init__(self, *args, default_max_tokens=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_max_tokens = default_max_tokens

    def __str__(self):
        return "OpenAI Completion: {}".format(self.model_id)

    def execute(self, prompt, stream, response, conversation=None):
        if prompt.system:
            raise NotImplementedError(
                "System prompts are not supported for OpenAI completion models"
            )
        messages = []
        if conversation is not None:
            for prev_response in conversation.responses:
                messages.append(prev_response.prompt.prompt)
                messages.append(prev_response.text())
        messages.append(prompt.prompt)
        response._prompt_json = {"messages": messages}
        kwargs = self.build_kwargs(prompt)
        if stream:
            completion = openai.Completion.create(
                model=self.model_name or self.model_id,
                prompt="\n".join(messages),
                stream=True,
                **kwargs,
            )
            chunks = []
            for chunk in completion:
                chunks.append(chunk)
                content = chunk["choices"][0].get("text") or ""
                if content is not None:
                    yield content
            response.response_json = combine_chunks(chunks)
        else:
            completion = openai.Completion.create(
                model=self.model_name or self.model_id,
                prompt="\n".join(messages),
                stream=False,
                **kwargs,
            )
            response.response_json = completion.to_dict_recursive()
            yield completion.choices[0]["text"]


def not_nulls(data) -> dict:
    return {key: value for key, value in data if value is not None}


def combine_chunks(chunks: List[dict]) -> dict:
    content = ""
    role = None
    finish_reason = None

    # If any of them have log probability, we're going to persist
    # those later on
    logprobs = []

    for item in chunks:
        for choice in item["choices"]:
            if (
                "logprobs" in choice
                and "text" in choice
                and isinstance(choice["logprobs"], dict)
                and "top_logprobs" in choice["logprobs"]
            ):
                logprobs.append(
                    {
                        "text": choice["text"],
                        "top_logprobs": choice["logprobs"]["top_logprobs"],
                    }
                )
            if "text" in choice and "delta" not in choice:
                content += choice["text"]
                continue
            if "role" in choice["delta"]:
                role = choice["delta"]["role"]
            if "content" in choice["delta"]:
                content += choice["delta"]["content"]
            if choice.get("finish_reason") is not None:
                finish_reason = choice["finish_reason"]

    # Imitations of the OpenAI API may be missing some of these fields
    combined = {
        "content": content,
        "role": role,
        "finish_reason": finish_reason,
    }
    if logprobs:
        combined["logprobs"] = logprobs
    for key in ("id", "object", "model", "created", "index"):
        if key in chunks[0]:
            combined[key] = chunks[0][key]

    return combined

PLUGIN_CLASS = OpenAIPlugin


from typing import List, Set, Optional, Iterator, Dict, Any
from dataclasses import dataclass, field
from pydantic import BaseModel
from sqlite_utils import Database
from abc import ABC, abstractmethod
from ulid import ULID
from datetime import datetime
import time
import json
from tango.configs.dbs_config import NANOLLM_DB

class ModelNotFoundError(Exception):
    pass

class UnknownModelError(KeyError):
    pass

@dataclass
class Prompt:
    prompt: str 
    model: "Model" 
    conversation: "Conversation" 
    id: str = field(default_factory=lambda: str(ULID()).lower())
    system: Optional[str] = None
    template: Optional[str] = None
    options: "Options" = field(default_factory=lambda: Options())

    def log_prompt(self):
        db = Database(NANOLLM_DB)
        db["prompts"].upsert({
            "id": self.id,
            "conversation_id": self.conversation.id,
            "timestamp": datetime.utcnow().isoformat(),
            "prompt": self.prompt,
            "system": self.system,
            "template": self.template
        }, pk="id")
    
    @classmethod
    def from_row(cls, row, conversation = None, model = None):
        if model is None:
            model = DummyModel.get_by_id(row["model_id"])
            if model is None:
                raise ModelNotFoundError(f"Model with ID {row['model_id']} not found.")
        if conversation is None:
            conversation = Conversation.get_by_id(row["conversation_id"])
            if conversation is None:
                raise ModelNotFoundError(f"Conversation with ID {row['conversation_id']} not found.")
        return cls(
            prompt=row["prompt"],
            model=model,
            conversation=conversation,
            system=row["system"],
            template=row["template"],
            id=row["id"]
        )
    @classmethod
    def get_by_id(cls, prompt_id, conversation= None, model = None):
        db = Database(NANOLLM_DB)
        prompt = db["prompts"].get(prompt_id)
        return cls.from_row(prompt, conversation, model) if prompt else None
@dataclass
class Conversation:
    model: "Model"
    id: str = field(default_factory=lambda: str(ULID()).lower())
    responses: List["Response"] = field(default_factory=list)

    def prompt(
        self,
        prompt: str = "",
        system: Optional[str] = None,
        stream: bool = True,
        **options
    ):
        return Response(
            Prompt(
                prompt=prompt,
                conversation=self,
                system=system,
                model=self.model,
                options=self.model.ModelOptions(**options),
            ),
            self.model,
            conversation=self,
            stream = stream
        )

    def start(self):
        db = Database(NANOLLM_DB)
        self.start_time = datetime.utcnow().isoformat()
        db["conversations"].upsert({
            "id": self.id,
            "start_time": self.start_time,
            "end_time": None,
            "model_id": self.model.model_id
        }, pk="id")

    def end(self):
        db = Database(NANOLLM_DB)
        self.end_time = datetime.utcnow().isoformat()
        db["conversations"].upsert({
            "id": self.id,
            "end_time": self.end_time
        }, pk="id")

    @classmethod
    def from_row(cls, row, model=None):
        if model is None:
            model = DummyModel.get_by_id(row["model_id"])
            if model is None:
                raise ModelNotFoundError(f"Model with ID {row['model_id']} not found.")
        return cls(
            model=model,
            id=row["id"])

    @classmethod
    def get_latest(cls):
        db = Database(NANOLLM_DB)
        rows = list(db["conversations"].rows_where(order_by="start_time"))
        latest_conv = rows[-1] if rows else None
        if latest_conv:
            conversation =  cls.from_row(latest_conv) 
            for response in db["responses"].rows_where( "conversation_id = ?", [conversation.id]):
              conversation.responses.append(Response.from_row(response, 
                                                              conversation=conversation, model=conversation.model))
            return conversation
        else:
            return None
    
    @classmethod
    def get_by_id(cls, conv_id, model=None):
        db = Database(NANOLLM_DB)
        conv = db["conversations"].get(conv_id)
        return cls.from_row(conv, model=model) if conv else None
    
class Response(ABC):
    def __init__(
        self,
        prompt: Prompt,
        model: "Model",
        conversation: Conversation,
        stream: bool,
    ):
        self.prompt = prompt
        self._prompt_json = None
        self.model = model
        self.stream = stream
        self._chunks: List[str] = []
        self._done = False
        self.response_json = None
        self.conversation = conversation

    def __iter__(self) -> Iterator[str]:
        self._start = time.monotonic()
        self._start_utcnow = datetime.utcnow()
        if self._done:
            return self._chunks
        for chunk in self.model.execute(
            self.prompt,
            stream=self.stream,
            response=self,
            conversation=self.conversation,
        ):
            yield chunk
            self._chunks.append(chunk)
        if self.conversation:
            self.conversation.responses.append(self)
        self._end = time.monotonic()
        self._done = True
        self.log_response()

    def _force(self):
        if not self._done:
            list(self)

    def text(self) -> str:
        self._force()
        return "".join(self._chunks)
    
    def result(self) -> str:
        return "".join(self._chunks)
    
    def json(self) -> Optional[Dict[str, Any]]:
        return self.response_json
    
    def duration_ms(self) -> int:
        return int((self._end - self._start) * 1000)

    def datetime_utc(self) -> str:
        return self._start_utcnow.isoformat()
    
    def log_response(self):
        db = Database(NANOLLM_DB)
        db["responses"].upsert({
            "id": str(ULID()).lower(),
            "prompt": self.prompt.prompt,
            "prompt_json": self._prompt_json,
            "system": self.prompt.system,
            "response": self.result(),
            "response_json": self.json(),
            "duraion_ms": self.duration_ms(),
            "start_datetime_utc": self.datetime_utc(),
            "timestamp": datetime.utcnow().isoformat(),
            "model_id": self.model.model_id,
            "prompt_id": self.prompt.id,
            "conversation_id": self.conversation.id,
            "start":self._start,
            "end": self._end,
        }, pk="id")

    @classmethod
    def from_row(cls, row, conversation = None, model = None):
        if model is None:
            model = DummyModel.get_by_id(row["model_id"])
            if model is None:
                raise ModelNotFoundError(f"Model with ID {row['model_id']} not found.")
        if conversation is None:
            conversation = Conversation.get_by_id(row["conversation_id"], model=model)
            if conversation is None:
                raise ModelNotFoundError(f"Conversation with ID {row['conversation_id']} not found.")
        prompt = Prompt.get_by_id(row["prompt_id"], conversation, model)
        if prompt is None:
            raise ModelNotFoundError(f"Prompt with ID {row['prompt_id']} not found.")
                 
        response = cls(
            model=model,
            prompt= prompt,
            conversation = conversation,
            stream=False,
        )
        response.id = row["id"]
        response._prompt_json = json.loads(row["prompt_json"] or "null")
        response.response_json = json.loads(row["response_json"] or "null")
        response._done = True
        response._chunks = [row["response"]]
        return response


class Options(BaseModel):
    # Note: using pydantic v1 style Configs,
    # these are also compatible with pydantic v2
    class Config:
        extra = "forbid"

        
class Model(ABC):
    model_id: str
    key: Optional[str] = None
    needs_key: Optional[str] = None
    key_env_var: Optional[str] = None
    can_stream: bool = False

    def __init__(self, model_id, key=None, needs_key=None, key_env_var=None, can_stream=False):
        self.model_id = model_id
        self.key = key
        self.needs_key = needs_key
        self.key_env_var = key_env_var
        self.can_stream = can_stream
        self._options: Dict[str, Any] = {}

    class ModelOptions(Options):
        pass

    @abstractmethod
    def execute(
        self,
        prompt: Prompt,
        stream: bool,
        response: Response,
        conversation: Optional[Conversation],
    ) -> Iterator[str]:
        """
        Execute a prompt and yield chunks of text, or yield a single big chunk.
        Any additional useful information about the execution should be assigned to the response.
        """
        pass

    def conversation(self):
        conv = Conversation(model=self)
        conv.start()  # Start and log the conversation
        return conv

    def prompt(self, conversation:Conversation, prompt_input: str, system: Optional[str] = None, stream: bool = True, **options):
        response = self.response(Prompt(prompt=prompt_input, system=system, conversation=conversation, model=self, 
                                        options=self.ModelOptions(**options)), conversation=conversation, stream=stream)
        response.prompt.log_prompt()  # Log the prompt
        return response

    def response(self, prompt: Prompt, conversation:Conversation, stream: bool = True) -> Response:
        resp = Response(prompt, self, conversation, stream)
        return resp

    def __str__(self) -> str:
        return "{}: {}".format(self.__class__.__name__, self.model_id)

    def __repr__(self):
        return "<Model '{}'>".format(self.model_id)


class DummyModel(Model):
    def execute(self, prompt: Prompt, stream: bool, response: Response, conversation: Conversation | None) -> Iterator[str]:
        yield "dummy response from dummy model"

    @classmethod
    def from_row(cls, row):
        # Return Model static information
        return cls(
            model_id=row["model_id"],
            key=row["key"],
            needs_key=row["needs_key"],
            key_env_var=row["key_env_var"],
            can_stream=row["can_stream"]
        )

    @classmethod
    def get_by_id(cls, model_id):
        db = Database(NANOLLM_DB)
        row = db["models"].get(model_id)
        return cls.from_row(row) if row else None

@dataclass
class ModelWithAliases:
    model: Model
    aliases: Set[str]

    @classmethod
    def from_row(cls, row):
        model = DummyModel.from_row(row)
        aliases = set(json.loads(row["aliases"]))
        return cls(model=model, aliases=aliases)

    @classmethod
    def load_dummy_models_with_aliases(cls):
        db = Database(NANOLLM_DB)
        rows = list(db["models"].rows)
        return {row['model_id']: cls.from_row(row) for row in rows}
    
    def log_model_with_aliases(self):
        db = Database(NANOLLM_DB)
        db["models"].upsert({
            "model_id": self.model.model_id,
            "key": self.model.key,
            "needs_key": self.model.needs_key,
            "key_env_var": self.model.key_env_var,
            "can_stream": self.model.can_stream,
            "aliases": json.dumps(list(self.aliases))
        }, pk="model_id")
    



import pluggy
import json
import os
from pathlib import Path
from typing import List, Set, Optional, Dict
from tango.nanoLLM.nanollm.plugin_spec import NanoLLMSpec, hookimpl
from tango.nanoLLM.nanollm.models import Model, ModelWithAliases, UnknownModelError
from nanollm import user_dir

class NanoLLMPluginManager:
    """Manages plugins for nanoLLM."""
    
    def __init__(self):
        self.pm = pluggy.PluginManager("nanoLLM")
        self.pm.add_hookspecs(NanoLLMSpec)
        self.pm.load_setuptools_entrypoints("nanoLLM")
        # Load models and aliases from disk with DummyModels
        self.models_and_aliases: Dict[str, ModelWithAliases] = ModelWithAliases.load_dummy_models_with_aliases() 
        self.default_model = None
        self.config:Dict = self.load_config()
    
    def load_config(self):
        config_path = user_dir() / "config.json"
        if config_path.exists():
            return json.loads(config_path.read_text())
        else:
            return {
                "default_model": None,
            }
    
    def set_default_model(self, model_id_or_alias) -> str:
        """Set the default model."""
        # Check if the model exists
        model = self.get_model_by_model_id_or_alias(model_id_or_alias)
        if model:
            self.config['default_model'] = model.model_id
            with open(user_dir() / "config.json", "w") as f:
                json.dump(self.config, f, indent=2)
            return model.model_id
        else:
            raise ValueError(f"No {model_id_or_alias} found.")

    def get_default_model(self):
        return self.config['default_model']

    def register_model_alias(self, model: Model, aliases: Set[str]=set()):
        """Register a model along with its aliases."""
        if model.model_id in self.models_and_aliases:
            self.models_and_aliases[model.model_id].aliases.update(aliases)
        else:
            model_with_aliases = ModelWithAliases(model=model, aliases=aliases)
            self.models_and_aliases[model.model_id] = model_with_aliases
        # Update dummy model with registered model
        self.models_and_aliases[model.model_id].model = model
        # Save the model with aliases to disk
        self.models_and_aliases[model.model_id].log_model_with_aliases()
    
    def set_model_alias(self, model_id_or_alias, alias):
        if model_id_or_alias in self.models_and_aliases:
            self.models_and_aliases[model_id_or_alias].aliases.add(alias)
        else:
            model = self.get_model_by_model_id_or_alias(model_id_or_alias)
            if model:
                self.register_model_alias(model, {alias})
            else:
                raise UnknownModelError(f"Model {model_id_or_alias} not found.")
        # Save the model with aliases to disk
        self.models_and_aliases[model_id_or_alias].log_model_with_aliases()

    def register_plugin(self, plugin, plugin_name):
        """Register a plugin."""
        self.pm.register(plugin, plugin_name)
        # Now, let's ask the plugin to load its models
        #plugin.register_models(self.register_model_alias)
    
    def get_model_by_model_id_or_alias(self, model_id_or_alias) -> Optional[Model]:
        """Retrieve a model by its name or alias."""
        if model_id_or_alias in self.models_and_aliases:
            return self.models_and_aliases[model_id_or_alias].model
        for model_with_aliases in self.models_and_aliases.values():
            if model_with_aliases.model.model_id == model_id_or_alias or model_id_or_alias in model_with_aliases.aliases:
                return model_with_aliases.model
        return None
    
    def list_models(self) -> List[Model]:
        return [model_with_aliases.model for model_with_aliases in self.models_and_aliases]

    # This method will allow plugins to register commands
    def hook(self):
        return self.pm.hook
"""
def get_models_with_aliases(pm) -> List["ModelWithAliases"]:
    model_aliases = []

    # Include aliases from aliases.json
    aliases_path = user_dir() / "aliases.json"
    extra_model_aliases: Dict[str, list] = {}
    if aliases_path.exists():
        configured_aliases = json.loads(aliases_path.read_text())
        for alias, model_id in configured_aliases.items():
            extra_model_aliases.setdefault(model_id, []).append(alias)

    def register(model, aliases=None):
        alias_list = list(aliases or [])
        if model.model_id in extra_model_aliases:
            alias_list.extend(extra_model_aliases[model.model_id])
        model_aliases.append(ModelWithAliases(model, alias_list))

    pm.hook().register_models(register=register)

    return model_aliases
"""


import logging
import sqlite3
import json
from tango.configs.dbs_config import NANOLLM_DB

def logs_db_path():
    #return user_dir() / "logs.db"
    return NANOLLM_DB

def _human_readable_size(size: int) -> str:
    """Convert a size in bytes to a human-readable format."""
    # Define the units and their respective byte sizes
    units = ["B", "KB", "MB", "GB", "TB", "PB"]
    unit_index = 0

    # Loop to find the appropriate unit for the size
    while size > 1024 and unit_index < len(units) - 1:
        size /= 1024
        unit_index += 1

    # Format the size to 2 decimal places and append the unit
    return "{:.2f} {}".format(size, units[unit_index])

def logs_db_size(db_path):
    return _human_readable_size(db_path.stat().st_size)


#logger=setup_logging()

class SQLiteHandler(logging.Handler):
    """Custom logging handler that writes logs to an SQLite database."""
    
    def __init__(self, db_path):
        super().__init__()
        self.db_path = db_path
        self._initialize_db()
        
    def _initialize_db(self):
        """Ensure the logs table exists in the database."""
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        
        # Create logs table if not exists
        cursor.execute('''
        CREATE TABLE IF NOT EXISTS logs (
            id INTEGER PRIMARY KEY,
            timestamp TEXT NOT NULL,
            event TEXT NOT NULL,
            message TEXT NOT NULL,
            level TEXT NOT NULL,
            other_data TEXT
        )
        ''')
        
        # Commit and close the connection
        conn.commit()
        conn.close()
        
    def emit(self, record):
        # Connect to the SQLite database
        conn = sqlite3.connect(self.db_path)
        cursor = conn.cursor()
        
        # Prepare structured log data
        timestamp = record.created
        event = record.name
        message = record.getMessage()
        level = record.levelname
        other_data = json.dumps(record.__dict__)
        
        # Insert log entry into the logs table
        cursor.execute('''
        INSERT INTO logs (timestamp, event, message, level, other_data)
        VALUES (?, ?, ?, ?, ?)
        ''', (timestamp, event, message, level, other_data))
        
        # Commit and close the connection
        conn.commit()
        conn.close()

# The SQLiteHandler now includes the method _initialize_db to ensure the logs table exists in the database.
# You can now use this updated class in your main script to handle logging to the SQLite database.


if __name__ == '__main__':
    # Configure logging to use the SQLiteHandler
    logging.basicConfig(level=logging.INFO)  # Log all events from INFO level and above
    logger = logging.getLogger("nanoLLM")
    logger.addHandler(SQLiteHandler(LOG_DB))
    
    # Test the logging configuration
    logger.info("Logging system initialized", extra={"event": "system_init"})
    
    "Logging configuration set up and test log entry created."
    
    # Fetch the latest log entries from the SQLite database
    # Connect to the SQLite database
    conn = sqlite3.connect(LOG_DB)
    cursor = conn.cursor()
    
    # Fetch the last 5 log entries
    cursor.execute('''
    SELECT timestamp, event, message, level, other_data 
    FROM logs 
    ORDER BY id DESC 
    LIMIT 5
    ''')
    log_entries = cursor.fetchall()
    
    # Close the connection
    conn.close()
    
    print(json.dumps(log_entries))

def setup_logger():
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger("nanoLLM")
    logger.addHandler(SQLiteHandler(NANOLLM_DB))
    #Avoid propagating the log messages to the root logger to prevent duplicate messages such as console output.
    logger.propagate = False
    return logger
logger = setup_logger()




import os
import sys
from pathlib import Path

def user_dir():
    llm_user_path = os.environ.get("NLLM_USER_PATH")
    if llm_user_path:
        path = Path(llm_user_path)
    else:
        path = Path(os.path.expanduser("~/.nanollm"))
    path.mkdir(exist_ok=True, parents=True)
    return path

def read_prompt(prompt):
    # Is there extra prompt available on stdin?
    stdin_prompt = None
    if not sys.stdin.isatty():
        stdin_prompt = sys.stdin.read()

    if stdin_prompt:
        bits = [stdin_prompt]
        if prompt:
            bits.append(prompt)
        prompt = " ".join(bits)

    if prompt is None and sys.stdin.isatty():
        # Hang waiting for input to stdin (unless --save)
        prompt = sys.stdin.read()
    return prompt


import importlib
import importlib.util
import os
from tango.nanoLLM.nanollm.plugin_manager import NanoLLMPluginManager

# Example usage
#pm = NanoLLMPluginManager()
#register_plugins(pm)

PLUGINS_DIR = os.path.join(os.path.dirname(__file__), 'plugins')

def discover_plugins():
    """Discover all plugins in the plugins directory."""
    plugins = []
    for item in os.listdir(PLUGINS_DIR):
        if item.endswith('.py') and item != '__init__.py':
            plugins.append(item[:-3])  # remove .py
    return plugins

def camel_case(string):
    """Convert snake_case string to CamelCase."""
    return ''.join(word.capitalize() for word in string.split('_'))

def register_plugins(pm:NanoLLMPluginManager):
    """Dynamically import and register plugins with the provided plugin manager."""
    plugins = discover_plugins()
    for plugin_name in plugins:
        try:
            # Construct full plugin module path
            module_path = f"tango.nanoLLM.nanollm.plugins.{plugin_name}"
            
            # Import plugin module dynamically
            plugin_module = importlib.import_module(module_path)

            # Check for PLUGIN_CLASS marker
            plugin_class = getattr(plugin_module, 'PLUGIN_CLASS', None)

            # If PLUGIN_CLASS does not exist, try camel_case approach
            plugin_class_name = camel_case(plugin_name)
            if not plugin_class:
                plugin_class = getattr(plugin_module, plugin_class_name, None)
            
            if plugin_class:
                # Instantiate the plugin class and register it
                pm.register_plugin(plugin_class(), plugin_name)
            else:
                print(f"Couldn't find the plugin class {plugin_class_name} in module {module_path}")
            
        except Exception as e:
            print(f"Failed to register plugin {plugin_name}. Error: {e}")
    
def setup_plugin_manager(pm):
    register_plugins(pm)
    pm.hook().register_models(register=pm.register_model_alias)


import click
import os
import json
import sqlite_utils
from click_default_group import DefaultGroup
from tango.nanoLLM.nanollm.plugin_manager import NanoLLMPluginManager
from tango.nanoLLM.nanollm.plugin_utils import setup_plugin_manager
from tango.nanoLLM.nanollm.templates import Template, template_dir, load_template
from tango.nanoLLM.nanollm.models import Conversation, ModelWithAliases
from tango.nanoLLM.nanollm.sqlite_logging import logger, logs_db_path, logs_db_size
from nanollm import (
    user_dir,
    read_prompt
)


pm = NanoLLMPluginManager()
setup_plugin_manager(pm)

@click.group(
    cls=DefaultGroup,
    default="prompt",
    default_if_no_args=True
    )
def cli():
    pass

@cli.group(name="keys")
@click.pass_context
def keys(ctx):
    "Commands for managing keys"
    ctx.ensure_object(dict)
   
@keys.command("set")
@click.argument('key')
@click.argument('value', required=False)
@click.pass_context
def set_key(ctx, key, value):
    """Set a key-value pair in the context object."""
    ctx.obj[key] = value
    click.echo(f"Key '{key}' set to '{value}'")

@keys.command(name="get")
@click.argument('key')
@click.pass_context
def get_key(ctx, key):
    """Retrieve the value for a given key from the context object."""
    value = ctx.obj.get(key)
    if value is not None:
        click.echo(value)
    else:
        click.echo(f"No value set for key '{key}'")

@cli.command()
@click.argument('prompt', default="", required=True)
@click.option("-s", "--system", help="System prompt to use")
@click.option('--model-id', '--model', '-m', default=pm.get_default_model(), help='Model to use for prediction')
@click.option("--stream", is_flag=True, help="Do stream output")
@click.option("--template", "-t", default=None, help="Name of the template to use")
@click.option("--param", "-p", multiple=True, type=(str, str), help="Parameters for the template in the form -p key value")
@click.option( "options", "-o", "--option", type=(str, str), multiple=True, help="key/value options for the model")
@click.option("-n", "--no-log", is_flag=True, help="Don't log to database")
@click.option(
    "continue_latest",
    "-c",
    "--continue",
    is_flag=True,
    flag_value=-1,
    help="Continue the most recent conversation.",
)
@click.option(
    "conv_id",
    "--cid",
    "--conversation",
    help="Continue the conversation with the given ID.",
)
def prompt(model_id, prompt, system, stream, template, param, options, no_log, continue_latest, conv_id):
    """Get a response for the given input prompt."""

    if model_id:
        # Fetch the model by name
        model = pm.get_model_by_model_id_or_alias(model_id)
        if not model:
            click.echo(f"No model found with name: {model}")
            return

        prompt = read_prompt(prompt)

        # Determine which conversation to use
        if continue_latest or conv_id:
            if continue_latest:
                conversation = Conversation.get_latest()
                if not conversation:
                    click.echo("No previous conversation found. Starting a new one.")
                    conversation = model.conversation()
            else:
                conversation = Conversation.get_by_id(conv_id)
                if not conversation:
                    click.echo(f"No conversation found with ID {conv_id}. Starting a new one.")
                    conversation = model.conversation()
        else:
            conversation = model.conversation()

        should_stream = model.can_stream and stream
        prompt_method = model.prompt
        
        # If a template is provided, generate the prompt using the template
        if template:
            tmpl = load_template(template)
            if not tmpl:
                click.echo(f"No template found with name: {template}")
                return
            params_dict = dict(param)  # Convert tuple of key-value pairs into a dictionary
            prompt, system = tmpl.evaluate(prompt, params_dict)
        
        options_dict = dict(options)

        response = prompt_method(conversation, prompt, system, **options_dict)
        if should_stream:
            # Stream the response
            for chunk in response:
                click.echo(chunk)
        else:
            click.echo(response.text())

        # Log the conversation
        if not no_log:
            logger.info(f"Model: {model_id}, Prompt: {prompt}, Stream: {stream}", 
                        extra={"conversation": conversation.id})

    else:
        click.echo("No model specified.")
    

@cli.group(name="models")
def models():
    "Commands for managing models"

@models.command(name='set-default')
@click.argument('model_id_or_alias')
def set_default_model_cmd(model_id_or_alias):
    """Set the default model."""
    model_id = pm.set_default_model(model_id_or_alias)
    click.echo(f"Default model set to: {model_id}")

@models.command(name='default')
def get_default_model_cmd():
    """Get the default model."""
    model = pm.get_default_model()
    if model:
        click.echo(model)
    else:
        click.echo("No default model is set.")

@models.command(name="list")
def list_models():
    """List all registered models."""
    for model_id, model_with_alias in pm.models_and_aliases.items():
        click.echo(f"Model: {model_id}")
        if model_with_alias.aliases:
            click.echo(f"Aliases: {', '.join(model_with_alias.aliases)}")

@models.command(name="set-alias")
@click.argument('model_id_or_alias', required=True)
@click.argument('alias', required=True)
def set_model_alias(model_id_or_alias, alias):
    """Set an alias for a model."""
    pm.set_model_alias(model_id_or_alias, alias)

@cli.group(
    cls=DefaultGroup,
    default="list",
    default_if_no_args=True,
)
def template():
    """Commands for managing templates."""
    pass

@template.command()
@click.option('--name', required=True, help="Name of the template.")
@click.option('--prompt', required=False, help="Content for the prompt template.")
@click.option('--system', default="", help="Content for the system prompt template.")
def create(name, prompt, system):
    """Create a new template."""
    path = template_dir() / f"{name}.yaml"
    if path.exists():
        raise click.ClickException(f"Template named {name} already exists in {path}.")
    new_template = Template(name=name, prompt=prompt, system=system)
    new_template.log_template()
    with open(path, "w") as f:
        f.write(new_template.json())
    click.echo(f"Template {name} created successfully in {path}.")

@template.command()
def list():
    """List all available templates."""
    template_directory = template_dir()
    for template_file in template_directory.glob("*.yaml"):
        template_name = template_file.stem
        tmpl = load_template(template_name)
        click.echo(f"{template_name}  system:{tmpl.system or 'N/A'}, prompt:{tmpl.prompt or 'N/A'}")

@template.command(name="path")
def templates_path():
    "Output the path to the templates directory"
    click.echo(template_dir())

@template.command()
@click.option('--name', required=True, help="Name of the template to update.")
@click.option('--prompt', help="Updated content for the prompt template.")
@click.option('--system', help="Updated content for the system prompt template.")
def update(name, prompt, system):
    """Update an existing template."""
    path = template_dir() / f"{name}.yaml"
    if not path.exists():
        raise click.ClickException(f"Template named {name} does not exist.")
    loaded_template = load_template(name)
    if prompt:
        loaded_template.prompt = prompt
    if system:
        loaded_template.system = system
    with open(path, "w") as f:
        f.write(loaded_template.json())
    click.echo(f"Template {name} updated successfully.")

@template.command()
@click.option('--name', required=True, help="Name of the template to delete.")
def delete(name):
    """Delete a template."""
    path = template_dir() / f"{name}.yaml"
    if not path.exists():
        raise click.ClickException(f"Template named {name} does not exist.")
    os.remove(path)
    click.echo(f"Template {name} deleted successfully.")

@cli.group()
def logs():
    """Commands for managing logs."""
    pass

@logs.command(name="path")
def logs_path():
    "Output the path to the logs.db file"
    click.echo(logs_db_path())


@logs.command(name="status")
def logs_status():
    "Show current status of database logging"
    path = logs_db_path()
    if not path.exists():
        click.echo("No log database found at {}".format(path))
        return
    db = sqlite_utils.Database(path)
    click.echo("Found log database at {}".format(path))
    #click.echo("Number of conversations logged:\t{}".format(db["conversations"].count))
    #click.echo("Number of responses logged:\t{}".format(db["responses"].count))
    click.echo("Number of logs logged:\t{}".format(db["logs"].count))
    click.echo(
        "Database file size: \t\t{}".format(logs_db_size(path))
    )
    for log in db["logs"].rows:
        click.echo(log)
        break

if __name__ == "__main__":
    pm.hook().register_commands(cli=cli)
    cli()



import os
import sys

class SuppressOutput:
    def __enter__(self):
        # Save a copy of the current file descriptors for stdout and stderr
        self.stdout_fd = os.dup(1)
        self.stderr_fd = os.dup(2)

        # Open a file to /dev/null
        self.devnull_fd = os.open(os.devnull, os.O_WRONLY)

        # Replace stdout and stderr with /dev/null
        os.dup2(self.devnull_fd, 1)
        os.dup2(self.devnull_fd, 2)

        # Writes to sys.stdout and sys.stderr should still work
        self.original_stdout = sys.stdout
        self.original_stderr = sys.stderr
        sys.stdout = os.fdopen(self.stdout_fd, "w")
        sys.stderr = os.fdopen(self.stderr_fd, "w")

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore stdout and stderr to their original state
        os.dup2(self.stdout_fd, 1)
        os.dup2(self.stderr_fd, 2)

        # Close the saved copies of the original stdout and stderr file descriptors
        os.close(self.stdout_fd)
        os.close(self.stderr_fd)

        # Close the file descriptor for /dev/null
        os.close(self.devnull_fd)

        # Restore sys.stdout and sys.stderr
        sys.stdout = self.original_stdout
        sys.stderr = self.original_stderr
